/*
 * vim: ts=4 nobackup
 */


#ifndef  __FUZZY_H__
#define  __FUZZY_H__


#include <stdint.h>
#include "Vector.H"


extern "C++" {
	namespace avelino {
		using namespace std;

		class FuzzySet {
			protected:
				float a, b, c, d;
			public:
				FuzzySet(float a, float b, float c, float d);
				float getMembership(float v);
				float getCenter();
		};

		class FuzzyValueReader {
			public:
				virtual float getValue() = 0;
		};

		class FuzzyExpression {
			public:
				virtual float getValue() = 0;
		};

		class FuzzyTerminalExpression : public FuzzyExpression {
			private:
				static Vector<FuzzyTerminalExpression *> pool;
			protected:
				FuzzyValueReader *vr;
				FuzzySet *s;
				FuzzyTerminalExpression(FuzzyValueReader &a, FuzzySet &b);
			public:
				static FuzzyTerminalExpression *getInstance(FuzzyValueReader &a, FuzzySet &b);
				float getValue();
				FuzzyValueReader *getValueReader();
				FuzzySet *getFuzzySet();
		};

		class FuzzyBinaryExpression : public FuzzyExpression {
			protected:
				FuzzyExpression *a, *b;
			public:
				FuzzyBinaryExpression(FuzzyExpression &e1, FuzzyExpression &e2);
				FuzzyExpression *getExpression1();
				FuzzyExpression *getExpression2();
		};

		class FuzzyAndExpression : public FuzzyBinaryExpression {
			private:
				static Vector<FuzzyAndExpression *> pool;
			protected:
				FuzzyAndExpression(FuzzyExpression &e1, FuzzyExpression &e2);
			public:
				static FuzzyAndExpression *getInstance(FuzzyExpression &e1, FuzzyExpression &e2);
				float getValue();
		};

		class FuzzyOrExpression : public FuzzyBinaryExpression {
			private:
				static Vector<FuzzyOrExpression *> pool;
			protected:
				FuzzyOrExpression(FuzzyExpression &e1, FuzzyExpression &e2);
			public:
				static FuzzyOrExpression *getInstance(FuzzyExpression &e1, FuzzyExpression &e2);
				float getValue();
		};

		class FuzzyNotExpression : public FuzzyExpression {
			private:
				static Vector<FuzzyNotExpression *> pool;
			protected:
				FuzzyExpression *e;
				FuzzyNotExpression(FuzzyExpression &a);
			public:
				static FuzzyNotExpression *getInstance(FuzzyExpression &e);
				float getValue();
				FuzzyExpression *getInnerExpression();
		};

		class FuzzyValueWriter {
			protected:
				float num, den;
			public:
				FuzzyValueWriter();
				void init();
				void addValue(float v, float center);
				void done();
				virtual void setValue(float v) = 0;
		};

		class FuzzyConsequent {
			protected:
				FuzzySet *s;
				FuzzyValueWriter *vw;
			public:
				FuzzyConsequent(FuzzyValueWriter &a, FuzzySet &b);
				void setValue(float v);
				FuzzyValueWriter *getWriter();
		};

		class FuzzyRule {
			public:
				FuzzyExpression *e;
				FuzzyConsequent *c;
				FuzzyRule(FuzzyExpression &e1, FuzzyConsequent &c1);
				FuzzyRule(FuzzyExpression *e1, FuzzyConsequent *c1);
				void run();
		};

		class FuzzyController {
			protected:
				Vector<FuzzyValueWriter *> writers;
				Vector<FuzzyRule *> rules;
				void addValueWriter(FuzzyValueWriter *vw);
			public:
				void addRule(FuzzyExpression &e, FuzzyConsequent &c);
				void addRule(FuzzyExpression *e, FuzzyConsequent *c);
				void addRule(FuzzyExpression &e, FuzzyConsequent *c);
				void addRule(FuzzyRule *rule);
				void addRule(FuzzyRule &rule);
				void run();
		};


		class ErrorValueReader : public FuzzyValueReader {
			public:
				virtual float getValue();
		};

		class SpeedValueWriter : public FuzzyValueWriter {
			public:
				virtual void setValue(float v);
		};

		FuzzyExpression &operator % (FuzzyValueReader &vr, FuzzySet &fs);
		FuzzyExpression &operator && (FuzzyExpression &e1, FuzzyExpression &e2);
		FuzzyExpression &operator || (FuzzyExpression &e1, FuzzyExpression &e2);
		FuzzyExpression &operator ! (FuzzyExpression &e);
	}
}


#endif  // __FUZZY_H__
