/*
 * vim: ts=4 nobackup
 */


#include "Fuzzy.H"
#include <stdlib.h>


using namespace std;
using namespace avelino;


FuzzySet::FuzzySet(float a, float b, float c, float d) {
	this->a = a;
	this->b = b;
	this->c = c;
	this->d = d;
}


float FuzzySet::getMembership(float v) {
	float ret = 0;
	if ((v >= this->b) && (v <= this->c))
		ret = 1;
	else if ((v > this->a) && (v < this->b))
		ret = ((float) (v - this->a)) / (this->b - this->a);
	else if ((v > this->c) && (v < this->d))
		ret = 1 - (((float) (v - this->c)) / (this->d - this->c));
	return ret;
}


float FuzzySet::getCenter() {
	return (this->b + this->c) / 2;
}


Vector<FuzzyTerminalExpression *> FuzzyTerminalExpression::pool;


FuzzyTerminalExpression *FuzzyTerminalExpression::getInstance(FuzzyValueReader &a, FuzzySet &b) {
	FuzzyTerminalExpression *ret = NULL;
	FuzzyTerminalExpression **p = pool.values;
	for (uint8_t i = 0; i < pool.size; i++, p++) {
		if ( ((*p)->getValueReader() == &a) && ((*p)->getFuzzySet() == &b) ) {
			ret = *p;
			break;
		}
	}
	if (ret == NULL) {
		ret = new FuzzyTerminalExpression(a, b);
		pool.add(ret);
	}
	return ret;
}


FuzzyTerminalExpression::FuzzyTerminalExpression(FuzzyValueReader &a, FuzzySet &b) {
	this->vr = &a;
	this->s = &b;
}


float FuzzyTerminalExpression::getValue() {
	return this->s->getMembership(this->vr->getValue());
}


FuzzyValueReader *FuzzyTerminalExpression::getValueReader() {
	return this->vr;
}


FuzzySet *FuzzyTerminalExpression::getFuzzySet() {
	return this->s;
}


FuzzyBinaryExpression::FuzzyBinaryExpression(FuzzyExpression &e1, FuzzyExpression &e2) : a(&e1), b(&e2) {
}


FuzzyExpression *FuzzyBinaryExpression::getExpression1() {
	return this->a;
}


FuzzyExpression *FuzzyBinaryExpression::getExpression2() {
	return this->b;
}


Vector<FuzzyAndExpression *> FuzzyAndExpression::pool;


FuzzyAndExpression *FuzzyAndExpression::getInstance(FuzzyExpression &e1, FuzzyExpression &e2) {
	FuzzyAndExpression *ret = NULL;
	FuzzyAndExpression **p = pool.values;
	for (uint8_t i = 0; i < pool.size; i++, p++) {
		if ( ((*p)->getExpression1() == &e1) && ((*p)->getExpression2() == &e2) ) {
			ret = *p;
			break;
		}
	}
	if (ret == NULL) {
		ret = new FuzzyAndExpression(e1, e2);
		pool.add(ret);
	}
	return ret;
}


FuzzyAndExpression::FuzzyAndExpression(FuzzyExpression &e1, FuzzyExpression &e2) : FuzzyBinaryExpression(e1, e2) {
}


float FuzzyAndExpression::getValue() {
	float v1 = this->a->getValue();
	float v2 = this->b->getValue();
	return (v1 < v2) ? v1 : v2;   // min
}


Vector<FuzzyOrExpression *> FuzzyOrExpression::pool;


FuzzyOrExpression *FuzzyOrExpression::getInstance(FuzzyExpression &e1, FuzzyExpression &e2) {
	FuzzyOrExpression *ret = NULL;
	FuzzyOrExpression **p = pool.values;
	for (uint8_t i = 0; i < pool.size; i++, p++) {
		if ( ((*p)->getExpression1() == &e1) && ((*p)->getExpression2() == &e2) ) {
			ret = *p;
			break;
		}
	}
	if (ret == NULL) {
		ret = new FuzzyOrExpression(e1, e2);
		pool.add(ret);
	}
	return ret;
}


FuzzyOrExpression::FuzzyOrExpression(FuzzyExpression &e1, FuzzyExpression &e2) : FuzzyBinaryExpression(e1, e2) {
}


float FuzzyOrExpression::getValue() {
	float v1 = this->a->getValue();
	float v2 = this->b->getValue();
	return (v1 < v2) ? v2 : v1;   // max
}


Vector<FuzzyNotExpression *> FuzzyNotExpression::pool;


FuzzyNotExpression *FuzzyNotExpression::getInstance(FuzzyExpression &e) {
	FuzzyNotExpression *ret = NULL;
	FuzzyNotExpression **p = pool.values;
	for (uint8_t i = 0; i < pool.size; i++, p++) {
		if ((*p)->getInnerExpression() == &e) {
			ret = *p;
			break;
		}
	}
	if (ret == NULL) {
		ret = new FuzzyNotExpression(e);
		pool.add(ret);
	}
	return ret;
}


FuzzyNotExpression::FuzzyNotExpression(FuzzyExpression &a) : e(&a) {
}


float FuzzyNotExpression::getValue() {
	return 1 - this->e->getValue();
}


FuzzyExpression *FuzzyNotExpression::getInnerExpression() {
	return this->e;
}


FuzzyValueWriter::FuzzyValueWriter() {
	this->init();
}


void FuzzyValueWriter::init() {
	this->num = 0;
	this->den = 0;
}


void FuzzyValueWriter::addValue(float v, float center) {
	this->num += v * center;
	this->den += v;
}


void FuzzyValueWriter::done() {
	if (this->den != 0) {
		float aux = (float) (this->num / this->den);
		this->setValue(aux);
	}
}


FuzzyConsequent::FuzzyConsequent(FuzzyValueWriter &a, FuzzySet &b) {
	this->s = &b;
	this->vw = &a;
}


void FuzzyConsequent::setValue(float v) {
	this->vw->addValue(v, this->s->getCenter());
}


FuzzyValueWriter *FuzzyConsequent::getWriter() {
	return this->vw;
}


FuzzyRule::FuzzyRule(FuzzyExpression &e1, FuzzyConsequent &c1) : e(&e1), c(&c1) {
}


FuzzyRule::FuzzyRule(FuzzyExpression *e1, FuzzyConsequent *c1) : e(e1), c(c1) {
}


void FuzzyRule::run() {
	float v = this->e->getValue();
	this->c->setValue(v);
}


void FuzzyController::addValueWriter(FuzzyValueWriter *vw) {
	if (!this->writers.exists(vw))
		this->writers.add(vw);
}


void FuzzyController::addRule(FuzzyExpression &e, FuzzyConsequent &c) {
	this->addRule(new FuzzyRule(e, c));
}


void FuzzyController::addRule(FuzzyExpression *e, FuzzyConsequent *c) {
	this->addRule(new FuzzyRule(e, c));
}


void FuzzyController::addRule(FuzzyExpression &e, FuzzyConsequent *c) {
	this->addRule(new FuzzyRule(&e, c));
}


void FuzzyController::addRule(FuzzyRule *rule) {
	this->rules.add(rule);
	this->addValueWriter(rule->c->getWriter());
}


void FuzzyController::addRule(FuzzyRule &rule) {
	this->addRule(&rule);
}


void FuzzyController::run() {
	// init all ValueWriters
	FuzzyValueWriter **p = this->writers.values;
	for (uint8_t i = 0; i < this->writers.size; i++, p++)
		(*p)->init();
	// run all rules
	FuzzyRule **q = this->rules.values;
	for (uint8_t i = 0; i < this->rules.size; i++, q++)
		(*q)->run();
	// done all ValueWriters
	p = this->writers.values;
	for (uint8_t i = 0; i < this->writers.size; i++, p++)
		(*p)->done();
}


FuzzyExpression &avelino::operator % (FuzzyValueReader &vr, FuzzySet &fs) {
	FuzzyExpression *ret = FuzzyTerminalExpression::getInstance(vr, fs);
	return *ret;
}


FuzzyExpression &avelino::operator && (FuzzyExpression &e1, FuzzyExpression &e2) {
	FuzzyExpression *ret = FuzzyAndExpression::getInstance(e1, e2);
	return *ret;
}


FuzzyExpression &avelino::operator || (FuzzyExpression &e1, FuzzyExpression &e2) {
	FuzzyExpression *ret = FuzzyOrExpression::getInstance(e1, e2);
	return *ret;
}


FuzzyExpression &avelino::operator ! (FuzzyExpression &e) {
	FuzzyExpression *ret = FuzzyNotExpression::getInstance(e);
	return *ret;
}


float ErrorValueReader::getValue() {
	return 30;
}


void SpeedValueWriter::setValue(float v) {
}


/*
int main() {
	FuzzySet negativeError(-10000, -10000, -100, 0);
	FuzzySet noError(-100, 0, 0, 100);
	FuzzySet positiveError(0, 100, 10000, 10000);
	FuzzySet lowSpeed(0, 0, 0, 1000);
	FuzzySet mediumSpeed(0, 1000, 1000, 2000);
	FuzzySet highSpeed(1000, 2000, 2000, 2000);
	ErrorValueReader error;
	SpeedValueWriter speed;
	FuzzyController c;
	c.addRule(error % negativeError, new FuzzyConsequent(speed, lowSpeed));
	c.addRule(error % noError, new FuzzyConsequent(speed, mediumSpeed));
	c.addRule(error % positiveError, new FuzzyConsequent(speed, lowSpeed));
	c.run();
	return 0;
}
*/
